/**
 * @author Dave Toth
 * @date Summer and Fall 2006
 * 
 * This policy dictates the client's behavior if the client uses the buffer 
 * none policy for downloading workunits.
 */

package edu.wpi.cs.toth.task_distribution_simulator;

class BufferNoneDownloadEarlyPolicy extends DownloadPolicy{

    ///////////////////////////////////////////////////////////////////////////
    // Constants
    ///////////////////////////////////////////////////////////////////////////

    ///////////////////////////////////////////////////////////////////////////
    // Variables
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Indicates whether the first workunit has been downloaded.
     */
    private boolean firstWorkunitDownloaded;


    ///////////////////////////////////////////////////////////////////////////
    // Methods
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Constructor.  
     */
    public BufferNoneDownloadEarlyPolicy() {
	firstWorkunitDownloaded = false;
	downloadFlag = false;
    }
    

    /**
     * Returns how many workunits need to be downloaded.
     * @param client - A reference to the client.  This allows us to get the
     *                 information we need to do calculations without having
     *                 to modify every policy if we want to add a new policy.
     * @return - Returns how many workunits need to be downloaded.
     */
    public int howManyWorkunitsToDownload(SimulatedClient client) {

	// There are no tasks in the system
	if ((client.getCurrentWorkunit() == null) && (client.getNumberOfBufferedTasks() == 0)) {
	    return 1;
	}
	// There's a task but it's time to download the next one.
	if (((client.getNumberOfBufferedTasks() == 0) && (client.getCurrentWorkunit().getProgressSaved() + client.getCurrentWorkunit().getRecentTimeTowardsCompletion() >= (.95 * client.getCurrentWorkunit().getTaskDuration()))) || (downloadFlag == true)) {
	    return 1;
	}
	else {
	    return 0;
	}
    }

    /**
     * Returns whether the client needs to download another workunit 
     * immediately.
     * @param client - A reference to the client.  This allows us to get the
     *                 information we need to do calculations without having
     *                 to modify every policy if we want to add a new policy.
     * @return  - Returns whether the client needs to download another 
     *            workunit immediately.
     */
    boolean needToDownload(SimulatedClient client) {
	// If there's nothing to work on
	if ((client.getCurrentWorkunit() == null) && (client.getNumberOfBufferedTasks() < 1)) {
	    return true;
	}
	// If we're 95% done with the current task and there's nothing buffered
	if (((client.getCurrentWorkunit() != null) && (client.getCurrentWorkunit().getProgressSaved() + client.getCurrentWorkunit().getRecentTimeTowardsCompletion() >= .95*client.getCurrentWorkunit().getTaskDuration()) && (client.getNumberOfBufferedTasks() < 1)) || (downloadFlag == true)) {
	    return true;
	}
	// Otherwise we don't need to download
	else {
	    return false;
	}
    }

    /**
     * Returns whether the client needs to download an additional task to
     * fill the buffer to the correct state.  This occurs on the the first
     * time the client tries to fill the buffer for certain buffering methods.
     * In particular, Buffer None and Buffer None Dwonload Early never need
     * to do this, but Buffer 1 and Buffer Multiple do need to do this.
     * @param client - A reference to the client.  This allows us to get the
     *                 information we need to do calculations without having
     *                 to modify every policy if we want to add a new policy.
     * @return - Returns whether the client needs to download one additional
     *           task to fill the buffer correctly for the given policy.
     */
    boolean needToDownloadAdditionalTask(SimulatedClient client) {
	return false;
    }

    /**
     * Returns the time required to download a workunit.
     * @param client - A reference to the client.  This allows us to get the
     *                 information we need to do calculations without having
     *                 to modify every policy if we want to add a new policy.
     * @return Returns the time required to download a workunit.
     */
    int getDownloadTime(SimulatedClient client) {
	int fileSize = client.getFileSize();
	int downloadSpeed = client.getDownloadSpeed();
	if (fileSize % downloadSpeed != 0) {
	    return (int) (fileSize/downloadSpeed) + 1;
	}
	else {
	    return (int) (fileSize/downloadSpeed);
	}
	//	return 0;
    }   

    /**
     * Returns whether a task is pipelined.  Pipelined tasks do not
     * take the time to download them away from the client.  They are
     * assumed to be downloaded as the CPU is working on its current
     * workunit because the effect on the CPU from downloading is 
     * minimal.
     * @param client - A reference to the client.  This allows us to get the
     *                 information we need to do calculations without having
     *                 to modify every policy if we want to add a new policy.
     * @return Returns whether a task is pipelined.
     */
    boolean isTaskPipelined(SimulatedClient client) {
	if (client.getCurrentWorkunit() != null) {
	    return true;
	}
	else {
	    return false;
	}
    }

    /**
     * Sets the state of the download flag with true indicating a download
     * has been initiated but not completed.
     * @param state - The new state of the flag.
     */
    void setDownloadFlag(boolean state) {
	downloadFlag = state;
    }

    /**
     * Returns the state of the download flag with true indicating a download
     * has been initiated but not completed.
     * @return state - Returns the state of the flag.
     */
    boolean getDownloadFlag() {
	return downloadFlag;
    }
}
