/**
 * @author Dave Toth
 * @date Summer and Fall 2006
 * 
 * This class simulates a volunteer computing client.  It takes a trace of
 * a computer's available, unavailable but on, and unavailable times and
 * determines how many work units the computer would have completed in the
 * time covered by the trace.
 */

package edu.wpi.cs.toth.task_distribution_simulator;

import java.io.*;
import java.util.Vector;

class SimulatedClient {

    ///////////////////////////////////////////////////////////////////////////
    // Constants
    ///////////////////////////////////////////////////////////////////////////

    // Task to do in next time interval.
    static final int CREATE_CHECKPOINT = 1;
    static final int RESTORE_FROM_CHECKPOINT = 2;
    static final int DOWNLOAD_WORKUNITS = 3;
    static final int PROCESS_WORKUNITS = 4;

    // The cost of checkpointing (in seconds).
    static final int CHECKPOINT_COST = 300;//10;
    static final int RESTORE_COST = 300;//10;

    // The time computation time between checkpoints (in seconds).
    public static final int CHECKPOINT_INTERVAL = 8620;//1574;

    // Computer status.
    static final String screenSaverOn = "AVAIL";
    static final String screenSaverOff = "ON-UN";
    static final String computerOff = "OFF";

    // Indicates whether 0) Check to see if we need to refill the work buffer
    //                or 1) We DO need to refill the work buffer.
    static final int IS_IT_TIME = 0;
    static final int REFILL = 1;


    ///////////////////////////////////////////////////////////////////////////
    // Variables
    ///////////////////////////////////////////////////////////////////////////

    /**
     * The file containing the computer trace to use.
     */
    private String dataFile;

    /**
     * The method of buffering tasks to use.
     */
    private String bufferingMethod;
  
    /**
     * The number of tasks to buffer.
     */
    private int tasksToBuffer;

    /**
     * The amount of time after a task has been downloaded before it is aborted if
     *  it has not been completed.
     */
    private int delayBound;

    /**
     * The amount of time a task should take.
     */
    private int taskDuration;

    /**
     * The bits per second that a client can download.
     */
    private int downloadSpeed;

    /**
     * The size of the file (in bits) to download for a workunit.
     */
    private int fileSize;
    
    /**
     * The percent of CPU time a client gets if the computer is unavailable 
     * but on.
     */
    private float percentCPU;

    /**
     * The time the trace covers and thus the time the simulation runs.
     */
    private int traceTime;

    /**
     * The array representing the event calendar.
     */
    private int[] eventCalendar;

    /**
     * Vector to hold downloaded but not started workunits.
     */
    private Vector toStartWorkunits;

    /**
     * Vector to hold downloaded and started workunits.
     */
    private Vector startedWorkunits;

    /**
     * The workunit being processed.
     */
    private Workunit currentWorkunit;

    /**
     * The CPU time available to spend on workunits.
     */
    private int availableTime;

    /**
     * Time spent on downloads that were halted mid-download due to 
     * powering off the system.
     */
    private int failedDownloadTime;

    /**
     * Time spent on downloads that were halted mid-download  
     * because the trace ended.
     */
    private int endFailedDownloadTime;

    /**
     * Indicates to the buffer multiple method to download 1 additional task
     * the first time to get the buffer in the right state.
     */
    private int firstTime;

    /**
     * The download policy the client will use.
     */
    private DownloadPolicy policy;

    /**
     * Indicates whether we need to 0: check to see if we need to refill the 
     * work buffer or 1: we DO need to refill the work buffer.
     */
    private int testOrRefill;

    /**
     * Tracks the the time during the simulation.
     */
    private int currentTime;


    ///////////////////////////////////////////////////////////////////////////
    // Methods
    ///////////////////////////////////////////////////////////////////////////

   
    /**
     * Constructor.  Runs the simulation with the command line arguments and 
     * outputs the results.
     * @param dataFile - The file containing the computer trace to use.
     * @param bufferingMethod - The method of buffering tasks to use.
     * @param daysOfTasksToBuffer - The number of days of tasks to buffer.
     * @param delayBound - The amount of time a task gets between being 
     *                     downloaded and aborted.
     * @param taskDuration - The amount of time a task should take.
     * @param downloadSpeed - The bits per second that a client can download.
     * @param fileSize - The bits in the file to download for a workunit. 
     * @param policy - The download policy the client will use.
     */
    public SimulatedClient(String dataFile, String bufferingMethod, float daysOfTasksToBuffer, int delayBound, int taskDuration, int downloadSpeed, int fileSize, DownloadPolicy policy) {
	this.dataFile = dataFile;
	this.bufferingMethod = bufferingMethod;
	this.delayBound = delayBound;
	this.taskDuration = taskDuration;
	this.downloadSpeed = downloadSpeed;
	this.fileSize = fileSize;
	this.policy = policy;

	this.tasksToBuffer = Math.max(1, (int) ((daysOfTasksToBuffer * 86400)/this.taskDuration));

	// Echo the parameters for debugging.
	/*System.out.println("\nParameters:\n==============");
	System.out.println("dataFile: " + this.dataFile);
	System.out.println("bufferingMethod: " + this.bufferingMethod);
	System.out.println("tasksToBuffer: " + this.tasksToBuffer);
	System.out.println("delayBound: " + this.delayBound);
	System.out.println("taskDuration: " + this.taskDuration);
	System.out.println("downloadSpeed: " + this.downloadSpeed);
	System.out.println("fileSize: " + this.fileSize);
	System.out.println("==============\n");*/
	//System.exit(0);

	if ((bufferingMethod.equals(Simulator.bufferingMethod1)) && (fileSize == 0)) {
	    System.err.print("\n" + this.dataFile);
	    System.err.print("  bufferingMethod: " + this.bufferingMethod);
	    System.err.print("," + this.tasksToBuffer);
	    System.err.print(", " + this.delayBound);
	    System.err.print(", " + this.taskDuration);
	    System.err.print(", " + this.downloadSpeed);
	    System.err.print(", " + this.fileSize + "--,");
	}	

	toStartWorkunits = new Vector();
	startedWorkunits = new Vector();

	failedDownloadTime = 0;
	endFailedDownloadTime = 0;

	// Keep track of how much time could have been spent on workunits.
	availableTime = 0;

	currentTime = 0;

	readTrace();
	runSimulation();
	writeStatistics();
    }


    /**
     * Reads the trace file into the array for the simulation.
     */
    private void readTrace() {
	// Open the input file and read in the trace, storing the data in 
	// the event calendar array.
	try {
	    File inputFile = new File(dataFile);
	    BufferedReader br = new BufferedReader(new FileReader(inputFile));
	    String traceData = br.readLine();
	    traceTime = Integer.parseInt(traceData.substring(traceData.indexOf(",") + 1));
	    /*System.out.println("Trace Time: " + traceTime + "\n");*/
	    //System.out.println("Intervals:\n----------");
	    eventCalendar = new int[traceTime];
	    //System.out.println("array size: " + eventCalendar.length);
	    int currentSlot = 0;

	    //////////////////////////////////////////////////////////////////////////
	    // WE'RE ASSUMING HERE THAT EVERYTHING IS DONE IN 1 SECOND
	    // INTERVALS.  THUS, IF THE TRACE FILE SAYS THE COMPUTER WAS IDLE
	    // FOR 30 SECONDS, WE FILL 30 ARRAY SLOTS WITH IDLE=100
	    //
	    // WE USE 1 SECOND BLOCKS BECAUSE SOME PERIOD DURATIONS ARE
	    // SURPRISINGLY NOT MULTIPLES OF 10.  FOR EXAMPLE, THERE IS AN
	    // OFF PERIOD OF 61 SECONDS.  THIS CAN RESULT FROM SCHEDULER ISSUES
	    // SINCE WINDOWS IS NOT REAL-TIME, SO MY SLEEP 10 SECONDS CALL MAY
	    // HAVE RESULTED IN SLEEPING FOR 11 SECONDS AND ALSO WHEN THE 
	    // COMPUTER GETS TURNED OFF, THE TIME IT'S TURNED BACK ON IS NOT 
	    // NECESSARILY SOME MULTIPLE OF TEN SECONDS FROM WHEN IT WAS TURNED
	    // OFF. 
	    //////////////////////////////////////////////////////////////////////////

	    traceData = br.readLine();
	    while (traceData.indexOf(",") >= 0) {
		// Process the line.
		/*System.out.print(traceData + " ");*/
		int tempDuration = Integer.parseInt(traceData.substring(traceData.indexOf(",") + 1));
		/*System.out.println("Interval Duration: " + tempDuration);*/
		for (int index = currentSlot; index < currentSlot + tempDuration; index ++) {
		    //System.out.println("curent slot: " + index);
		    if (traceData.indexOf(screenSaverOn) >= 0) {
			eventCalendar[index] = 100;
			availableTime = availableTime + 1;
		    }
		    else {
			if (traceData.indexOf(screenSaverOff) >= 0) {
			    eventCalendar[index] = 5;
			    //availableTime = availableTime + percentCPU;
			}
			else {
			    eventCalendar[index] = 0;
			}
		    }
		}
		currentSlot = currentSlot + tempDuration;

		// Grab the next line.
		traceData = br.readLine();
	    }

	    /*
	    // DEBUG
	    System.out.println("DEBUGGING");
	    
	    for (int debugger = 0; debugger < eventCalendar.length; debugger ++) {
		System.out.println("time[" + debugger+ "]=" + eventCalendar[debugger]);
	    }
	    
	    // END DEBUG
	    */

	}
	catch (Exception e) {
	    System.out.println(e.toString());
	    System.out.println("There was an error reading the trace file: " + dataFile);
	    System.exit(0);
	}
    }


    /**
     * Write the results of the simulation and the parameters to a file.
     */
    private void writeStatistics() {

	//System.out.println("Writing Stats");

	// The file to write the results into.
	File outputFile = null;
	
	// The BufferedWriter to use to write the output file.
	BufferedWriter bw = null;
	
	// The FileWriter to use to write the output file.
	FileWriter fw = null;
	
	// Create the output file and write the simulation parameters to it.

	/*
	String fileName = new String("Results_" + bufferingMethod + "_" + Integer.toString(tasksToBuffer) + "_" + Integer.toString(delayBound) + "_" + Integer.toString(taskDuration) + "_" + Integer.toString(downloadSpeed) + "_" + Integer.toString(fileSize) + "_" + dataFile + ".dmt");
	outputFile = new File(fileName);
	
	try {
	    fw = new FileWriter(outputFile);
	    bw = new BufferedWriter(fw);
	    bw.write(this.dataFile);
	    bw.newLine();
	    bw.write(this.bufferingMethod);
	    bw.newLine();
	    bw.write("" + this.tasksToBuffer);
	    bw.newLine();
	    bw.write("" + this.delayBound);
	    bw.newLine();
	    bw.write("" + this.taskDuration);
	    bw.newLine();
	    bw.write("" + this.downloadSpeed);
	    bw.newLine();
	    bw.write("" + this.fileSize);
	    bw.newLine();
	}
	catch (IOException ioe) {
	    System.out.println("Writing simulation parameters to the output file failed.");
	    System.exit(0);
	}
	*/

	// Write the statistics out to the file.
	//try {	

	    // The number of aborted workunits.
	    int abortedWorkunits = 0;
	    
	    // The number of completed workunits.
	    int completedWorkunits = 0;
	    
	    // The number of workunits that have not been resolved.
	    int unresolvedWorkunits = 0;
	    
	    // The amount of time spent downloading and working on completed workunits.
	    int timeOnCompletedWorkunits = 0;
	    
	    // The amount of time spent downloading and working on aborted workunits.
	    int timeOnAbortedWorkunits = 0;
	    
	    // The amount of time spent downloading and working on unresolved workunits.
	    int timeOnUnresolvedWorkunits = 0;

	    //System.out.println("\n\n-----");

	    // Step through units in completed vector.
	    if (startedWorkunits.size() <=0 ) { // If there are no workunits that have been started we have a huge problem.  Exit.
		System.err.print(",");
		//System.out.println("SIMULATION FAILED - NO WORKUNITS EVER STARTED");
		//bw.write("SIMULATION FAILED - NO WORKUNITS EVER STARTED");
		//bw.newLine();
		//bw.flush();
		//fw.close();
		System.exit(0);
	    }
	    else {
		failedDownloadTime = endFailedDownloadTime;
		while (startedWorkunits.size() > 0) {
		    Workunit tempUnit = (Workunit) startedWorkunits.remove(0);
		    failedDownloadTime = failedDownloadTime + tempUnit.getWastedDownloadTime();
		    //System.out.print("Workunit: " + tempUnit.getWorkunitID() + " ");
		    if (tempUnit.getCompletionStatus() == Workunit.ON_TIME) { // Workunit was completed on time.
			completedWorkunits ++;
			//System.out.println("Completed ON TIME");
			timeOnCompletedWorkunits = timeOnCompletedWorkunits + tempUnit.getTimeUsed();
		    }
		    else {
			if (tempUnit.getCompletionStatus() == Workunit.LATE) { // Workunit was late and aborted.
			    abortedWorkunits ++;
			    //System.out.println("ABORTED");
			    timeOnAbortedWorkunits = timeOnAbortedWorkunits + tempUnit.getTimeUsed();
			}
			else { // Workunit was not resolved - there was insufficient trace time to resolve it.
			    timeOnUnresolvedWorkunits = timeOnUnresolvedWorkunits + tempUnit.getTimeUsed();
			    //System.out.println("%%%%%%%%%%%%%%%%%% " + timeOnUnresolvedWorkunits);
			    //System.out.println("UNRESOLVED");
			    unresolvedWorkunits ++;
			}
		    }
		    /*System.out.println("Total Time: " + tempUnit.getTimeUsed());
		    System.out.println("Wasted Download Time: " + tempUnit.getWastedDownloadTime()); 
		    System.out.println("-----");
		    System.out.println(timeOnUnresolvedWorkunits);*/
		}
		while (toStartWorkunits.size() > 0) {
		    Workunit tempUnit = (Workunit) toStartWorkunits.remove(0);	
		    //System.out.print("Workunit: " + tempUnit.getWorkunitID() + " ");
		    failedDownloadTime = failedDownloadTime + tempUnit.getWastedDownloadTime();
		    if (tempUnit.getCompletionStatus() == Workunit.ON_TIME) { // Workunit was completed on time.
			completedWorkunits ++;
			//System.out.println("Completed ON TIME");
			timeOnCompletedWorkunits = timeOnCompletedWorkunits + tempUnit.getTimeUsed();
		    }
		    else {
			if (tempUnit.getCompletionStatus() == Workunit.LATE) { // Workunit was late and aborted.
			    abortedWorkunits ++;
			    //System.out.println("ABORTED");
			    timeOnAbortedWorkunits = timeOnAbortedWorkunits + tempUnit.getTimeUsed();
			}
			else { // Workunit was not resolved - there was insufficient trace time to resolve it.
			    timeOnUnresolvedWorkunits = timeOnUnresolvedWorkunits + tempUnit.getTimeUsed();
			    //System.out.println("UNRESOLVED");
			    unresolvedWorkunits ++;
			    if (fileSize % downloadSpeed != 0) {
				//timeOnUnresolvedWorkunits = timeOnUnresolvedWorkunits + (int) (fileSize/downloadSpeed) + 1;
				//System.out.println("############################## " + timeOnUnresolvedWorkunits);
			    }
			    else {
				//timeOnUnresolvedWorkunits = timeOnUnresolvedWorkunits + (int) (fileSize/downloadSpeed);
				//System.out.println("############################## " + timeOnUnresolvedWorkunits);
			    }
			}
		    }
		    //System.out.println("Total Time: " + tempUnit.getTimeUsed());
		    //System.out.println("Wasted Download Time: " + tempUnit.getWastedDownloadTime()); 
		    // System.out.println("-----");
		    //System.out.println(timeOnUnresolvedWorkunits);
		}
	    }

	    // Step through units in to do vector.

	    // Print number of units completed on time, number failed, number unresolved.
	    // Print trace time, time spent on units completed on time, time spent
	    // on units failed, time spent on units unresolved.
	    //bw.newLine();
	    //bw.newLine();
	    System.err.print(completedWorkunits + ",");
	    //bw.write("Completed Workunits: " + completedWorkunits);
	    //bw.newLine();
	    //System.err.print(timeOnCompletedWorkunits + ",");
	    //bw.write("Time Spent On Completed Workunits: " + timeOnCompletedWorkunits);
	    //bw.newLine();
	    //bw.newLine();
	    //System.err.print(abortedWorkunits + ",");
	    //System.err.print(timeOnAbortedWorkunits + ",");
	    if ((bufferingMethod.equals(Simulator.bufferingMethod3)) && (tasksToBuffer == 14)) {System.err.println(" ");}
	    //bw.write("Aborted Workunits: " + abortedWorkunits);
	    //bw.newLine();
	    //bw.write("Time Spent On Aborted Workunits: " + timeOnAbortedWorkunits);
	    //bw.newLine();
	    //bw.newLine();
	    //bw.write("Unresolved Workunits: " + unresolvedWorkunits);
	    //bw.newLine();
	    //bw.write("Time Spent On Unresolved Workunits: " + timeOnUnresolvedWorkunits);
	    //bw.newLine();
	    //bw.newLine();
	    //System.out.println("Trace Time: " + traceTime);
	    //bw.write("Trace Time: " + traceTime);
	    //bw.newLine();
	    //System.out.println("Total Available Time To Spend on Workunits: " + availableTime);
	    //bw.write("Total Available Time To Spend on Workunits: " + availableTime);
	    //bw.newLine();
	    //bw.write("Time spent on downloads interrupted and thus not completed: " + failedDownloadTime);
	    //bw.newLine();

	    //bw.flush();
	    //fw.close();
	    //}
	    //catch (IOException ioe) {
	    //    System.out.println("Writing statistics to the output file failed.");
	    //    System.exit(0);
	    //}
    }




    /**
     * Returns whether there's enough time to complete an operation before the computer is powered off.
     * @param currentTime - The time to start looking at.
     * @param timeNeeded - How much time is needed to complete the operation.
     * @return - Returns the time the operation would finish or (minus 1 * the time the operation would fail)
     */
    private int enoughTimeToCompleteOperation(int currentTime, int timeNeeded) {
	// We scan forward through the array, counting the number of positions with 1's 
	// until we find the first 0 or we are sure there's enough time to complete the
	// operation.
	
	//System.out.println("ENOUGH TIME METHOD: " + currentTime);
	int theTime = currentTime;
	int counter = 0;
	while ((counter < timeNeeded) && (theTime < traceTime) && (eventCalendar[theTime] != 0)) {
	    if (eventCalendar[theTime] == 100) {
		counter ++;
	    }
	    theTime ++;
	}

	if (counter >= timeNeeded) {
	    //System.out.println("returning: " + theTime);
	    return theTime;
	}
	else {
	    //System.out.println("returning: " + (theTime*-1));
	    return (theTime * -1);
	}
    }


    /**
     * Returns the abort multiple.
     * @return - Returns the Delay Bound.
     */
    int getDelayBound() {
	return delayBound;
    }

    /**
     * Returns the task duration.
     * @return - Returns the task duration.
     */
    int getTaskDuration() {
	return taskDuration;
    }

    /**
     * Returns a sring with the name of the data file being used in the 
     * simulation.
     * @return - Returns a sring with the name of the data file being used in
     *           the simulation.
     */
    String getTraceFileName() {
	return dataFile;
    }

    /**
     * Returns the current time of the simulation.
     * @return - Returns the current time of the simulation.
     */
    int getCurrentTime() {
	return currentTime;
    }

    /**
     * Returns the number of tasks in the client's task buffer.
     * @return - Returns the number of tasks in the client's task buffer.
     */
    int getNumberOfBufferedTasks() {
	return toStartWorkunits.size();
    }

    /**
     * Returns whether we need to test to see if we need to refill the work
     * buffer or whether we DO need to refill the work buffer. 
     * @return - Returns whether we need to test to see if we need to refill
     *           the work buffer or whether we DO need to refill the work 
     *           buffer. 
     */
    int getDownloadStatus() {
	return testOrRefill;
    }

    /**
     * Returns the number of tasks the client will buffer.
     * @return - Returns  the number of tasks the client will buffer.
     */
    int getNumberOfTasksToBuffer() {
	return tasksToBuffer;
    }

    /**
     * Returns whether this is the first time we're filling the buffer
     * with the Buffer Multiple policy.
     * @return - Returns whether this is the first time we're filling
     *           the buffer with the Buffer Multiple policy.
     */
    int isFirstTimeFillingBuffer() {
	return firstTime;
    }

    /**
     * Returns the current workunit being processed or null if none
     * is being processed.
     * @return - Returns the current workunit being processed or null
     *           if none is being processed.
     */
    Workunit getCurrentWorkunit() {
	if (currentWorkunit != null) {
	    return currentWorkunit;
	}
	else {
	    return null;
	}
    }

    /** Returns the size of the file needed to do a workunit in bits.
     * @return - Returns the size of the file needed to do a workunit in bits.
     */
    int getFileSize() {
	return fileSize;
    }

    /** Returns the client's download speed in bit/sec.
     * @return - Returns the client's download speed in bits/sec.
     */
    int getDownloadSpeed() {
	return downloadSpeed;
    }

    /**
     * Runs the simulation.
     */
    private void runSimulation() {
	
	// The number of workunits created thus far.  Let's count from 1 to n.
	int workUnitsInSimulation = 1;

	// We need to roll back the clock after certain downloads to give the
	// effect of some downloads being pipelined and thus "free" because
	// they happen while we're working on a task.  So we count up the 
	// time spent on downloads that should be pipelined and then when
	// that's donw, we set the clock back that amount and the normal
	// operations of create checkpoint, restore from checkpoint, and
	// process workunits use the time.
	int rollBackTime = 0;

	// The part of the rollback time left to use.
	int rollBackTimeLeft = -1;

	// Are we in the process of processing the time rolled back.
	boolean duringRollback = false;

	// The state we were in before we rolled back
	int activityBeforeRollback = PROCESS_WORKUNITS;

	//System.out.println("\n\nStarting Simulation\n\n");

	firstTime = 1;

	/**
	 * The amount of time that was wasted downloading a workunit because the 
	 * download was interrupted and had to be done again.
	 */
	int wastedDownloadTime = 0;

	// The activity to do in the next usable time interval.
	// This is initialized to download since we need to do that
	// before anything else when we start the client.
	int currentActivity = DOWNLOAD_WORKUNITS;

	// Indicates a download failed while a workunit was being processed, so the workunit
	// must be restarted from the last checkpoint.
	boolean restartFlag = false;

	// Go through the process of downloads and runs, accumulating the statistics.

	//System.out.println(currentTime + " " + traceTime);

	while (currentTime < traceTime) {

	    // If the CPU is available, use it.  
	    if (eventCalendar[currentTime] == 100) {
		
		// Select the appropriate activity.
		if ((restartFlag == true) /*&& (currentActivity != DOWNLOAD_WORKUNITS)*/ && (duringRollback == false)) {
		    //System.out.println("cur act: " + currentActivity);
		    currentActivity = RESTORE_FROM_CHECKPOINT;
		    restartFlag = false;
		}

		if (currentActivity == CREATE_CHECKPOINT) { // CREATE A CHECKPOINT
		    // If we have the time to create a checkpoint
		    int operationCompletionTime = enoughTimeToCompleteOperation(currentTime, CHECKPOINT_COST);
		    if (operationCompletionTime > 0) {
			// Create the checkpoint.
			currentWorkunit.successfullyCreateCheckpoint();

			// Advance time to compensate for the creation.
			currentTime = currentTime + CHECKPOINT_COST;
			//System.out.println("CHECKPOINT CREATION FINISHED: " + currentTime + " --- current progress saved: " + currentWorkunit.getProgressSaved() /*+ "       USED TIME: " + currentWorkunit.getTimeUsed()*/);

			
			// Set the state to what we should do next.
			currentActivity = PROCESS_WORKUNITS;

			rollBackTimeLeft = rollBackTimeLeft - CHECKPOINT_COST;
			if (rollBackTimeLeft <= 0) {
			    duringRollback = false;
			}
		    }
		    // We don't have the time to create a checkpoint
		    else {
			
			// Charge for the time we would waste trying to create a checkpoint.
			// First we need to make sure that we only charge for time that's usable.
			// Thus, we count the AVAILABLE units between when we start checkpointing and fail.
			int myTime = currentTime;
			int myCounter = 0;
			while ((myTime < (-1 * operationCompletionTime)) && (myTime < traceTime) && (eventCalendar[myTime] != 0)) {
			    if (eventCalendar[myTime] == 100) {
				myCounter ++;
			    }
			    myTime ++;
			}
			int wastedTime = myCounter;

			// Fail to create the checkpoint.
			currentWorkunit.failToCreateCheckpoint(wastedTime);
			//System.out.println("FAILED TO CREATE CHECKPOINT AT: " + (currentTime + wastedTime)); 

			// Advance the time to when we'd fail.
			currentTime = (-1 * operationCompletionTime);

			// Set the state to what we should do next.
			currentActivity = RESTORE_FROM_CHECKPOINT;

			rollBackTimeLeft = rollBackTimeLeft - myCounter;
			if (rollBackTimeLeft <= 0) {
			    duringRollback = false;
			}
		    }
		}

		else if ((currentActivity == RESTORE_FROM_CHECKPOINT) && (currentWorkunit != null)) { // RESTORE THE STATE OF A WORKUNIT FROM A CHECKPOINT
		    
		    // If we have the time to restore the state from the last checkpoint
		    int operationCompletionTime = enoughTimeToCompleteOperation(currentTime, RESTORE_COST);
		    if (operationCompletionTime > 0) {

			// Restore the workunit's state from the previous checkpoint.
			currentWorkunit.successfullyRestoreFromCheckpoint();

			// Set the state to what we should do next.
			currentActivity = PROCESS_WORKUNITS;

			// Advance the time to when we finish restoring.
			currentTime = currentTime + RESTORE_COST;
			//System.out.println("CHECKPOINT RESTORE FINISHED: " + currentTime);

			rollBackTimeLeft = rollBackTimeLeft - RESTORE_COST;
			if (rollBackTimeLeft <= 0) {
			    duringRollback = false;
			}
		    }
		    // We don't have the time to restore from the last checkpoint
		    else {

			// Charge for the time we would waste trying to restore from a checkpoint.
			// First we need to make sure that we only charge for time that's usable.
			// Thus, we count the AVAILABLE units between when we start restoring and fail.
			int myTime = currentTime;
			int myCounter = 0;
			while ((myTime < (-1 * operationCompletionTime)) && (myTime < traceTime) && (eventCalendar[myTime] != 0)) {
			    if (eventCalendar[myTime] == 100) {
				myCounter ++;
			    }
			    myTime ++;
			}
			
			rollBackTimeLeft = rollBackTimeLeft - myCounter;
			if (rollBackTimeLeft <= 0) {
			    duringRollback = false;
			}
			// Fail to restore from the previous checkpoint.
			currentWorkunit.failToRestoreFromCheckpoint(myCounter);

			//System.out.println("FAILED TO RESTORE FROM CHECKPOINT AT: " + (currentTime + myCounter));

			// Advance the time to when we'd fail.
			currentTime = (-1 * operationCompletionTime);

			// Set the state to what we should do next.
			currentActivity = RESTORE_FROM_CHECKPOINT;
		    }
		}

		else if ((currentActivity == DOWNLOAD_WORKUNITS) && (duringRollback == false)) { // DOWNLOAD WORKUNITS
		    // If you need to download workunit(s), do it.
		    testOrRefill = REFILL;
		    int workunitsToDownload = policy.howManyWorkunitsToDownload(this);
		    //System.out.println("DL TIME " + workunitsToDownload);

		    if (currentWorkunit != null) {
			/*System.out.println(currentWorkunit.getProgressSaved());*/
		    }
		    if (workunitsToDownload > 0) {

			//System.out.println("begin download at : " + currentTime);
			policy.setDownloadFlag(true);

			// Find out how long to download 1 workunit
			int timeToDownload = policy.getDownloadTime(this);

			// If you have time to download it.
			int operationCompletionTime = enoughTimeToCompleteOperation(currentTime, timeToDownload);
			if (operationCompletionTime >= 0) {
			    if (policy.isTaskPipelined(this) == true) {
				rollBackTime = rollBackTime + timeToDownload;
			    }
			    Workunit task;
			    if (policy.needToDownloadAdditionalTask(this) == true) {
				//System.out.println("first time");
				task = new Workunit(currentTime, taskDuration, delayBound, wastedDownloadTime, timeToDownload, toStartWorkunits.size(), workUnitsInSimulation, operationCompletionTime);
				task.setRequiredCompletionTime(task.getTimeReadyToProcess() + getDelayBound());
				workUnitsInSimulation ++;
			    }
			    else {
				//System.out.println("NOT first time");
				task = new Workunit(currentTime, taskDuration, delayBound, wastedDownloadTime, timeToDownload, toStartWorkunits.size() + 1, workUnitsInSimulation, operationCompletionTime);
				task.setRequiredCompletionTime(task.getTimeReadyToProcess() + getDelayBound());
				workUnitsInSimulation ++;
			    }
			    // Stuff in work queue.
			    toStartWorkunits.add(task);
			    //System.out.println("queued wu - buffered wu's: " + toStartWorkunits.size());

			    // Advance the time to the end of the download.
			    currentTime = operationCompletionTime;
			    //System.out.println("DOWNLOAD COMPLETED: " + currentTime + " ----- wasted time: " + wastedDownloadTime + "   WU#:" + task.getWorkunitID());

			    policy.setDownloadFlag(false);

			    // See if there are more workunits to download.
			    workunitsToDownload --;
			    if (workunitsToDownload > 0) {
				currentActivity = DOWNLOAD_WORKUNITS;
				//System.out.println("should be downloading next***********************");
			    }
			    else {
				currentActivity = PROCESS_WORKUNITS;
				//System.out.println("should be PROC next***********************");
				firstTime = 0;
				if (policy.isTaskPipelined(this) == true) {
				    currentTime = currentTime - rollBackTime;
				    rollBackTimeLeft = rollBackTime;
				    rollBackTime = 0;
				    duringRollback = true;
				}
			    }
			    wastedDownloadTime = 0;
			}
			// Otherwise, charge & fail.
			else {
			    // Set the state to what we should do next.
			    currentActivity = DOWNLOAD_WORKUNITS;
				
			    // First we need to make sure that we only charge for time that's usable.
			    // Thus, we count the AVAILABLE units between when we start downloading and fail.
			    int myTime = currentTime;
			    int myCounter = 0;
			    while ((myTime < (-1 * operationCompletionTime)) && (myTime < traceTime) && (eventCalendar[myTime] != 0)) {
				if (eventCalendar[myTime] == 100) {
				    myCounter ++;
				}
				myTime ++;
			    }
			    
			    if (policy.isTaskPipelined(this) == true) {
				rollBackTime = rollBackTime + myCounter;
			    }

			    // Advance the time to when we'd fail.
			    failedDownloadTime = myCounter; // /*failedDownloadTime +*/ ((-1 * operationCompletionTime) - currentTime); 

			    currentTime = (-1 * operationCompletionTime);
			    
			    /*System.out.println("            failed download time " + failedDownloadTime + " " + myCounter);*/
			    wastedDownloadTime = wastedDownloadTime + failedDownloadTime;

			    // Make sure that if we were in the middle of a workunit that we keep the accounting correct.
			    if (currentWorkunit != null) {			
				currentWorkunit.setRecentTimeTowardsCompletion(0);
				restartFlag = true;
			    }
			    // If the current workunit IS null, then we were not working on one, just downloading, so don't do anything.

			    /*System.out.println("TOTAL FAiled dowlnoaad at time: " + currentTime + "  --   wasted time: " + wastedDownloadTime);*/
			    //System.out.println("WU FAiled dowlnoaad at time: " + currentTime + "  --   wasted time: " + failedDownloadTime);
			    //System.out.println("Curr Act: " + currentActivity);

			    // If the last thing in the simulation is a failed download, then we need to save the failed time.
			    if (currentTime >= traceTime) {
				endFailedDownloadTime = failedDownloadTime;
			    }
			    //System.out.println("current act end proc: " + currentActivity);
			    if (policy.isTaskPipelined(this) == true) {
				currentTime = currentTime - rollBackTime;
				rollBackTimeLeft = rollBackTime;
				rollBackTime = 0;
				duringRollback = true;
			    }
			}
		    }
		    else {
			if ((workunitsToDownload == 0) && (bufferingMethod.equals(Simulator.bufferingMethod3))) {
			    //System.out.println("Trace: " + getTraceFileName() + " couldn't play nice");
			    //System.err.println("Trace: " + getTraceFileName() + " couldn't play nice");
			    //System.exit(0);
			    currentActivity = PROCESS_WORKUNITS;
			    /*System.out.println("-> TIME TO WORK");*/
			}
		    }
		}

		else { // WORK ON WORKUNITS
		    // First make sure we have a task to operate on.
		    if (currentWorkunit == null) {
			if (toStartWorkunits.size() < 1) {
			    currentActivity = DOWNLOAD_WORKUNITS;
			    /*System.out.println("--------------------------------DOWNLOADING DUE TO NULL&EMPTY");*/
			}
			else {
			    currentWorkunit = (Workunit) toStartWorkunits.remove(0);
			}
		    }
	    
		    // If we need to download a task, skip everything else so we can download a task.
		    boolean needToDownload =  policy.needToDownload(this);

		    //if (!((toStartWorkunits.size() < 1) && (currentWorkunit == null))) {
		    if ((needToDownload == true) && (duringRollback == false)) {
			/*System.out.println("===============================================need to dl at " + currentTime + " " + duringRollback);*/
			currentActivity = DOWNLOAD_WORKUNITS;	
			/*System.out.println("TIME TO DL ->");*/
		    }
		    else {
			// If the workunit is not ready because it's download was not completed yet, skip everything.
			if ((currentWorkunit == null) || (currentWorkunit.getTimeReadyToProcess() > currentTime)) {
			    // Do nothing.
			    //System.out.println("SPIN");
			}
			else {
			    // If the task will be aborted during this time unit.
			    if (currentWorkunit.getRequiredCompletionTime() < currentTime + 1) {
				
				// Abort the workunit.
				currentWorkunit.abort();
				startedWorkunits.add(currentWorkunit);
				currentWorkunit = null;
				//System.out.println("Workunit aborted: " + (currentTime + 1));
				//System.out.println("queued wu - buffered wu's: " + toStartWorkunits.size());
			    }
			    // The task won't be aborted during this time unit.
			    else { 
				
				currentWorkunit.setRecentTimeTowardsCompletion(currentWorkunit.getRecentTimeTowardsCompletion() + 1);
				
				// If the task will be completed during this interval.
				if (currentWorkunit.getRecentTimeTowardsCompletion() + currentWorkunit.getProgressSaved() >= currentWorkunit.getTaskDuration()) { 
				    
				    // We finish the task...
				    currentWorkunit.complete();
				    startedWorkunits.add(currentWorkunit);
				    //System.out.println("FINAL TALLY: " + currentWorkunit.getTimeUsed());
				    currentWorkunit = null;
				    
				    // ... and see what the state will be next unit of time
				    
			 	    // If we need to download workunits, the state next unit will be download
				    testOrRefill = IS_IT_TIME;
				    int workunitsToDownload = policy.howManyWorkunitsToDownload(this);
				    if (workunitsToDownload > 0) {
					/*System.out.println("POLICY 1 DICATES DL $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$ at " + currentTime);*/
					currentActivity = DOWNLOAD_WORKUNITS;
				    }
				    // Otherwise, the state next unit will be working on tasks
				    else {
					currentActivity = PROCESS_WORKUNITS;
				    }
				    
				    //System.out.println("Workunit completed: " + (currentTime + 1));
				    //System.out.println("queued wu - buffered wu's: " + toStartWorkunits.size());
				}
				// The task will not be completed during this interval.
				else { 
				    currentWorkunit.setTimeUsed(currentWorkunit.getTimeUsed() + 1);
				    
				    // Determine what we should do next interval.  
				    testOrRefill = IS_IT_TIME;
				    int workunitsToDownload = policy.howManyWorkunitsToDownload(this);
				    
				    // If we need to create a checkpoint
				    if (currentWorkunit.needToCheckpoint() == true) {
					currentActivity = CREATE_CHECKPOINT;
				    }
				    // If we need to download workunits, the state next unit will be download
				    else if (workunitsToDownload > 0) {
					/*System.out.println("POLICY 2 DICATES DL $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$");*/
					currentActivity = DOWNLOAD_WORKUNITS;
				    }
				    // Otherwise, the state next unit will be working on tasks
				    else {
					currentActivity = PROCESS_WORKUNITS;
				    }
				}
			    }
			}
			currentTime ++;
			rollBackTimeLeft --;
			if (rollBackTimeLeft <= 0) {
			    duringRollback = false;
			}
		    } // END if (!((toStartWorkunits.size() < 1) && (currentWorkunit == null)))
		}// END WORK ON WORKUNITS
		    
	    } // END if (eventCalendar[currentTime] == 100)
	    else {
		if (eventCalendar[currentTime] == 0) {
		    // Set the state to what we should do next.
		    // If we were in the middle of a download, restart the download
		    if (currentActivity == DOWNLOAD_WORKUNITS) {
			if (currentWorkunit != null) {
			    currentWorkunit.setRecentTimeTowardsCompletion(0);
			    restartFlag = true;
			}
		    }
		    else if (currentWorkunit == null) {
			currentActivity = PROCESS_WORKUNITS;
		    }
		    else {
			currentActivity = RESTORE_FROM_CHECKPOINT;
			currentWorkunit.setRecentTimeTowardsCompletion(0);
		    }
		}
		currentTime++;
	    }
	} // END while (currentTime < traceTime)
	if (currentWorkunit != null) {
	    // Make sure that the workunit is truly unresolved - it could be clearly that it gets aborted
	    // as soon as it starts up.
	    if (currentWorkunit.getRequiredCompletionTime() < currentTime + 1) {
		//currentWorkunit.setTimeUsed(currentWorkunit.getTimeUsed() + 1);
		currentWorkunit.setCompletionStatus(Workunit.LATE);
		//System.out.println("Workunit would be aborted at: " + (currentTime + 1));
	    }
	    startedWorkunits.add(currentWorkunit);
	}
    }
}
