/**
 * @author Dave Toth
 * @date Summer and Fall 2006
 * 
 * This class simulates a workunit for a volunteer computing project.  It
 * keeps track of the statistics for itself, such as time required to 
 * download the workunit, etc.
 */

package edu.wpi.cs.toth.task_distribution_simulator;

public class Workunit {

    ///////////////////////////////////////////////////////////////////////////
    // Constants
    ///////////////////////////////////////////////////////////////////////////

    // Workunit completion status.
    public static final int ON_TIME= 0;
    public static final int LATE = 1;
    public static final int UNDETERMINED = 2;


    ///////////////////////////////////////////////////////////////////////////
    // Variables
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Was the workunit completed on time, late, or undetermined (the 
     * simulation ended before this could be resolved).
     */
    private int completionStatus;
    
    /**
     * The CPU time (in seconds) spent on the workunit.
     */
    private int timeUsed;
    
    /**
     * The time (in seconds) spent on the workunit since the last checkpoint
     * that will actually count towards its competion.
     */
    private int recentTimeTowardsCompletion;
    
    /**
     * The time the workunit must be completed by (in seconds).  
     */
    private int requiredCompletionTime;
    
    /**
     * The time this workunit was assigned to the client (in seconds).
     */
    private int timeAssigned;
    
    /**
     * Task time that won't need to be repeated thanks to checkpointing.
     */
    private int progressSaved;
    
    /**
     * The time it takes to complete a task.
     */
    private int taskDuration;
    
    /**
     * The time spent downlaoding the workunit that didn't succeed so we had
     * to re-download it.
     */
    private int wastedDownloadTime;
    
    /**
     * The amount of time it takes to download this workunits if there are no
     * interruptions.
     */
    private int timeToDownload;
    
    /**
     * The unique ID of a workunit.  This number says whether the workunit was
     * created first (1), second (2), third (3), and so on.
     */
    private int ID;

    /**
     * The time the task is ready to be processed (i.e. download completed.
     */
    private int downloadCompletedAt;


    ///////////////////////////////////////////////////////////////////////////
    // Methods
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Constructor.  Creates a workunit and initializes its variables.
     * @param timeAssigned - The time the workunit was assigned to the client.
     * @param taskDuration - The CPU time (in seconds it takes to complete a 
     *                       workunit. 
     * @param delayBound - The amount of time a task gets after being downloaded 
     *                     before it's aborted.
     * @param wastedDownloadTime - The amount of time wasted downloading a 
     *                             workunit because the download was 
     *                             interrupted before it completes.
     * @param timeToDownload - The amount of time it takes to download this
     *                         workunit if there are no interruptions.
     * @param tasksBeforeMe - The number of tasks ahead of this one in the 
     *                        buffer.
     * @param ID - The ID number of the workunit.
     * @param timeReadyToProcess - The time the workunit's download was completed.
     */
    public Workunit(int timeAssigned, int taskDuration, float delayBound, int wastedDownloadTime, int timeToDownload, int tasksBeforeMe, int ID, int timeReadyToProcess) {
	this.timeAssigned = timeAssigned;
	this.taskDuration = taskDuration;
	this.wastedDownloadTime = wastedDownloadTime;
	this.timeToDownload = timeToDownload;
	this.ID = ID;
	downloadCompletedAt = timeReadyToProcess;
	
	// By default, the completionStatus should be UNDETERMINED.
	completionStatus = UNDETERMINED;
	
	// By default, the timeUsed should be 0;
	timeUsed = 0;
	
	progressSaved = 0;
	
	recentTimeTowardsCompletion = 0;
	
	timeUsed = wastedDownloadTime + this.timeToDownload;
	/*System.out.println("WU Created with wasted time: " + this.wastedDownloadTime);*/
    }
    
    
    /**
     * Returns whether the client needs to create a checkpoint for the current
     * workunit.
     * @return  - Returns whether the client needs to create a checkpoint for
     *            the current workunit.
     */
    boolean needToCheckpoint() {
	if (recentTimeTowardsCompletion >= SimulatedClient.CHECKPOINT_INTERVAL) {
	    return true;
	}
	else {
	    return false;
	}
    }

    /**
     * Sets the time that this workunit must be completed by.
     * @param time - The time the workunit must be completed by.  It
     *               is aborted the following second if it is not
     *               completed by that time.
     */
    void setRequiredCompletionTime(int time) {
	this.requiredCompletionTime = time;
	/*System.out.println("Task Created: " + timeAssigned + "   Abort at: " + requiredCompletionTime);*/
    }

    /**
     * Returns the unique identifier of the workunit, which is an integer 
     * indicating the order in which the workunit was created.
     * @return - Returns the unique identifier of the workunit, which is an 
     *           integer indicating the order in which the workunit was 
     *           created.
     */
    int getWorkunitID() {
	return ID;
    }

    /**
     * Returns the completion status of the workunit.
     * @return - Returns the completion status of the workunit.
     */
    int getCompletionStatus() {
	return completionStatus;
    }

    /**
     * Sets the completion status of a workunit.
     * @param status - The new completion status of the workunit.
     */
    void setCompletionStatus(int status) {
	completionStatus = status;
    }

    /**
     * Returns the time used to download and process this workunit.
     * @return - Returns the time used to download and process this workunit.
     */
    int getTimeUsed() {
	return timeUsed;
    }

    /**
     * Sets the time used by the workunit.
     * @param newValue - The new total amount of time the workunit has used
     *                   being downloaded and processed.
     */
    void setTimeUsed(int newValue) {
	timeUsed = newValue;
    }

    /**
     * Returns the amount of time wasted on failed attempts to download this
     * workunit.
     * @return Returns the amount of time wasted on failed attempts to 
     *         download this workunit.
     */
    int getWastedDownloadTime() {
	return wastedDownloadTime;
    }

    /**
     * Returns the amount of time (in seconds) spent on the workunit since 
     * the last checkpoint that will actually count towards its competion.
     * @return Returns the amount of time (in seconds) spent on the workunit 
     *         since the last checkpoint that will actually count towards
     *         its competion.
     */
    int getRecentTimeTowardsCompletion() {
	return recentTimeTowardsCompletion;
    }

    /**
     * Sets the amount of time (in seconds) spent on the workunit since 
     * the last checkpoint that will actually count towards its competion.
     * @param newValue - The new amount of time spent on the workunit since 
     *                   the last checkpoint that will actually count 
     *                   towards its competion.
     */
    void setRecentTimeTowardsCompletion(int newValue) {
	recentTimeTowardsCompletion = newValue;
    }

    /**
     * Returns the task time that won't need to be repeated thanks to 
     * checkpointing.
     * @return - Returns the task time that won't need to be repeated thanks to
     *           checkpointing.
     */
    int getProgressSaved() {
	return progressSaved;
    }

    /**
     * Sets the task time that won't need to be repeated thanks to 
     * checkpointing.
     * @param newValue - The new amount of task time that won't need to be 
     *                   repeated thanks to checkpointing.
     */
    void setProgressSaved(int newValue) {
	progressSaved = newValue;
    }

    /**
     * Returns the number of seconds a task will take to complete, not including
     * time to download the task.
     * @return - Returns the number of seconds a task will take to complete, not
     *           including time to download the task.
     */
    int getTaskDuration() {
	return taskDuration;
    }

    /**
     * Returns the time the workunit must be completed by (in seconds). 
     * @return -  Returns the time the workunit must be completed by (in 
     *            seconds).
     */
    int getRequiredCompletionTime() {
	return requiredCompletionTime;
    }

    /**
     * Abort the current workunit since it's now officially too late.
     */
    void abort() {
	setTimeUsed(getTimeUsed() + 1);
	setCompletionStatus(LATE);
    }

    /**
     * Complete the workunit.
     */
    void complete() {
	setTimeUsed(getTimeUsed() + 1);
	setCompletionStatus(ON_TIME);
    }

    /**
     * Successfully creates a checkpoint.
     */
    void successfullyCreateCheckpoint() {
	// Charge for the time we spent creating a checkpoint.
	setTimeUsed(getTimeUsed() + SimulatedClient.CHECKPOINT_COST);
	// Create the checkpoint.
	setProgressSaved(getProgressSaved() + SimulatedClient.CHECKPOINT_INTERVAL);
	setRecentTimeTowardsCompletion(0);
    }

    /**
     * Try to create a checkpoint but fail.
     * @param wastedTime - The usable time that was used trying to create the 
     *                     checkpoint before the failure occurred.
     */
    void failToCreateCheckpoint(int wastedTime) {
	setTimeUsed(getTimeUsed() + wastedTime);
	setRecentTimeTowardsCompletion(0);
    }

    /**
     *
     * @param 
     */
    void successfullyRestoreFromCheckpoint() {
	setTimeUsed(getTimeUsed() + SimulatedClient.RESTORE_COST);
	setRecentTimeTowardsCompletion(0);
    }

    /**
     * Try to restore from a checkpoint and fail.
     * @param wastedTime - The usable time that was used trying to restore
     *                     from the checkpoint before the failure occurred.
     */
    void failToRestoreFromCheckpoint(int wastedTime) {
	setTimeUsed(getTimeUsed() + wastedTime);
	setRecentTimeTowardsCompletion(0);
    }

    /**
     * Returns when the workunit can start being processed.
     * @return Returns when the workunit can start being processed.
     */
    int getTimeReadyToProcess() {
	return downloadCompletedAt;
    }
}
